#!/bin/bash

##
## NOTE: this particular version is tailormade for SLES15.
##


declare -A cfg
declare -r AOD_CFG_FILE="aod.cfg"
declare -r AOD_CFG_FILE_STRIPPED="aod_sanitized.cfg"
declare -r CFG_STRIP_SEDCMD='/(^#(.*)$|`(.*)`|\$\((.*)\)|eval|exec)/d'
declare -r CFG_VALIDKEYS_SEDCMD='/('\
'AOD_TMPDIR|'\
'AOD_INSTALLROOT|'\
'AOD_PERSISTENT_LOGDIR|'\
'AOD_TMPFS_SZ|'\
'AOD_NFS_PORT|'\
'AOD_TCPDUMP_BIN|'\
'AOD_TCPDUMP_BUFSZ_KB|'\
'AOD_TCPDUMP_SNAPLEN_BYTES|'\
'AOD_TCPDUMP_PCAPSZ_MB|'\
'AOD_TCPDUMP_MAX_PCAPS|'\
'AOD_LOG_LOOP_SLEEP_SEC|'\
'AOD_NUM_LOGS_RETAINED|'\
'AOD_NW_PROBE_ENABLE|'\
'AOD_NW_PROBE_TIMEOUT_SEC|'\
'AOD_NW_CAPTURE_ENABLE).*=/p'

declare -r CFG_TO_DICT_SEDCMD='s'\
'/\s*([A-Za-z0-9_]+)\s*=\s*([A-Za-z0-9_"]+)\s*'\
'/cfg[\U\1\E]=\2'\
'/'

declare -r NFS4_MOUNT_LINE_SEDCMD='s/(.+):(.+) on (.+) type nfs4.*\,addr=(.+).*\)/\1;\2;\3;\4/'

declare LOG_DISK_CLEANCMD

parse_cfg_file() {
    local cfg_file_path="";
    local cfg_file_sanitized="";
    local paths_to_check=(
        "."
        ".."
        "/opt/aod/"
    )

    for _path in ${paths_to_check[@]}
    do
        if [[ -r "${_path}/${AOD_CFG_FILE}" ]];
        then
            cfg_file_path="${_path}/${AOD_CFG_FILE}"
            cfg_file_sanitized="${_path}/${AOD_CFG_FILE_STRIPPED}"
            break
        fi
    done

    if [[ -z "$cfg_file_path" ]];
    then
        logger -i -p 'DAEMON.ERR' "aod: $cfg_file_path not readable, exiting."
        exit 1
    fi

    #
    # Note: eventually, we will have to move away from a bash script prototype
    # to something written in a first-class programming language. Until such
    # time, we should lock down the config file as much as possible. At the
    # moment, this boils down to (i) _only_ accepting a known set of config
    # keys, and (ii) stripping out any line in the config file that remotely
    # smells of containing executable code.
    #

    sed -En "${CFG_STRIP_SEDCMD}; ${CFG_VALIDKEYS_SEDCMD}" $cfg_file_path > $cfg_file_sanitized
    sed -Ei "${CFG_TO_DICT_SEDCMD}" $cfg_file_sanitized
    source $cfg_file_sanitized
}

populate_derived_cfg() {
    cfg[AOD_INSTALLDIR]="${cfg[AOD_INSTALLROOT]}/aod"
    cfg[AOD_TMPFS_MNTPT]="${cfg[AOD_INSTALLDIR]}/tmpfs_staging_area"
    cfg[AOD_STATIC_LOGDIR]="${cfg[AOD_PERSISTENT_LOGDIR]}/aod_static_info"
    cfg[AOD_ERR_LOG]="${cfg[AOD_PERSISTENT_LOGDIR]}/aod_err.log"
    cfg[AOD_PIDFILE]="${cfg[AOD_TMPDIR]}/aod.pidfile"
    cfg[AOD_TCPDUMP_PIDFILE]="${cfg[AOD_TMPDIR]}/aod_tcpdump.pidfile"
    cfg[AOD_TCPDUMP_LOGFILE]="${cfg[AOD_TMPDIR]}/aod_tcpdump.logfile"
    cfg[AOD_TCPDUMP_PCAP_PATTERN]="${cfg[AOD_TMPFS_MNTPT]}/nfs-pcap-"
    cfg[AOD_STOP_CAPTURE_MARKER]="${cfg[AOD_TMPDIR]}/aod.stop_capture"
    cfg[AOD_NW_PROBE_LOG]="${cfg[AOD_TMPFS_MNTPT]}/aod_nw_probe.out"
}

dump_current_cfg() {
    local key;
    local value;

    echo "$0: dumping current configuration"
    for key in ${!cfg[@]};
    do
        value=${cfg["$key"]}
        echo "$key : $value"
    done
}

log_err_to_syslog() {
    local errmsg="$1"
    local logline="[ $(date -u) ] $errmsg"
    logger -i -p 'DAEMON.ERR' "aod: $errmsg"
    echo $logline >>"${cfg[AOD_ERR_LOG]}"
}

log_alert_to_syslog() {
    local errmsg="$1"
    local logline="[ $(date -u) ] $errmsg"
    logger -i -p 'DAEMON.ALERT' "aod: ALERT: $errmsg"
    echo $logline >> ${cfg[AOD_ERR_LOG]}
}

mk_logdir_name() {
    local _lprfx;

    _lprfx="$(hostname)_$(date -u +'%Y_%m_%d_%H_%M_%S')_UTC_aod";
    echo "$_lprfx";
}

check_root() {
    local my_euid;

    my_euid=$(id -u);
    if (( my_euid != 0 ))
    then
        local errmsg="aod: Please run $0 as root, or under sudo. Exiting."
        logger -i -p 'DAEMON.ERR' "$errmsg";
        echo "$errmsg"
        exit 1;
    fi
}

is_valid_persistent_dir_path() {
    if [ "${cfg[AOD_PERSISTENT_LOGDIR]}" == "/" ] || [ "${cfg[AOD_PERSISTENT_LOGDIR]}" == "" ];
    then 
        return 0
    else
        return 1
    fi
}

check_persistent_dir_path() {
    is_valid_persistent_dir_path
    retval=$?
    echo $retval
    if [ $retval == 0 ];
    then
        local errmsg="aod: Please use a valid value other than "/" for AOD_PERSISTENT_LOGDIR. Exiting"
        logger -i -p 'DAEMON.ERR' "$errmsg";
        echo "$errmsg"
        exit 1;
    fi
}

check_other_running_instances() {
    local pid_from_pidfile=0
    local mypid="$$"
    local pidcount;

    if [[ -f ${cfg[AOD_STOP_CAPTURE_MARKER]} ]];
    then
        rm -f ${cfg[AOD_STOP_CAPTURE_MARKER]};
    fi

    if [[ -f ${cfg[AOD_PIDFILE]} ]];
    then
        pid_from_pidfile=$(cat ${cfg[AOD_PIDFILE]})

        if (( mypid != pid_from_pidfile ))
        then
            pidcount=$(ps -h -p "$pid_from_pidfile" |wc -l)
            if (( pidcount != 0 ))
            then
                log_err_to_syslog "$0: Another instance (PID $pid_from_pidfile) is already running. Exiting."
                exit 1
            fi
        fi
    fi

    echo $$ > ${cfg[AOD_PIDFILE]}
}

#check_resources() {
#    #
#    # [ ] Check /proc/meminfo, see how much memory we have, and how much we're
#    #     consuming now.
#    #
#    # [ ] Check df -lh, ensure that we have sufficient room on the disk where
#    #     /opt is mounted.
#    #
#    # [ ] Check the number of cores, make sure that we have _at least_ 3-4
#    #     cores. If not, bail out.
#    #
#    # [ ] Check the number of network interfaces. Discount `lo`. Look through
#    #     /proc/net/dev.
#    #
#    # [ ] Check to see if LSO/TSO is enabled on all/any of these interfaces.
#    #
#    # [ ] Do a quick back-of-the envelope calculation to see if we have enough
#    #     RAM and disk space. If not, what do we do? Do we still run with a
#    #     warning, or should we bail out?
#    #
#}

setup_disk_clean_job() {
    LOG_DISK_CLEANCMD="find ${cfg[AOD_PERSISTENT_LOGDIR]} -maxdepth 1 -type f -name \"*aod*.tgz\" -cmin +$((60*24)) -exec rm -f '{}' \;"

    # Cron job runs at midnight everyday
    local job="0 0 * * * ${LOG_DISK_CLEANCMD}"
    
    # Edit the crontab for the current user i.e. root
    ( crontab -l | grep -v -F "${LOG_DISK_CLEANCMD}" ; echo "$job" ) | crontab -
}

remove_disk_clean_job() {
    ( crontab -l | grep -v -F "${LOG_DISK_CLEANCMD}" ) | crontab -
}

mount_tmpfs() {
    local tmpfs_fstab_line="tmpfs ${cfg[AOD_TMPFS_MNTPT]}  tmpfs   size=${cfg[AOD_TMPFS_SZ]}    0   0"
    if (( $(mount -t tmpfs |grep -Ec "${cfg[AOD_TMPFS_MNTPT]}") == 0 ))
    then
        if (( $(grep -Ec "${cfg[AOD_TMPFS_MNTPT]}" /etc/fstab) == 0 ))
        then
            echo $tmpfs_fstab_line >> /etc/fstab
        fi

        mount "${cfg[AOD_TMPFS_MNTPT]}"
        if (( $? != 0 ))
        then
            echo "$0: error mounting tmpfs on ${cfg[AOD_TMPFS_MNTPT]}"
            exit 1
        fi
    fi
}

setup_tmp_logdirs() {
    local ret;

    if [[ ! -d "${cfg[AOD_TMPFS_MNTPT]}" ]]
    then
        mkdir -p "${cfg[AOD_TMPFS_MNTPT]}"
        mount_tmpfs
    fi

    if [[ ! -d "${cfg[AOD_PERSISTENT_LOGDIR]}" ]]
    then
        mkdir "${cfg[AOD_PERSISTENT_LOGDIR]}"
        if [[ ! -d "${cfg[AOD_PERSISTENT_LOGDIR]}" ]];
        then
            log_err_to_syslog "mkdir ${cfg[AOD_PERSISTENT_LOGDIR]} failed. Exiting."
            exit 1
        fi
    fi

    touch "${cfg[AOD_ERR_LOG]}"

    if [[ ! -d "${cfg[AOD_STATIC_LOGDIR]}" ]]
    then
        mkdir "${cfg[AOD_STATIC_LOGDIR]}"
        if [[ ! -d "${cfg[AOD_STATIC_LOGDIR]}" ]];
        then
            log_err_to_syslog "mkdir ${cfg[AOD_STATIC_LOGDIR]} failed [$ret]. Exiting."
            exit 1
        fi
    fi
}

collect_static_info() {
    ##  - lshw |awk '/-network/,/size/' |awk '/size/ {print $2}'
    ## NFS debug stuff?

    cat /etc/os-release > "${cfg[AOD_STATIC_LOGDIR]}/os-release"
    cat /proc/meminfo > "${cfg[AOD_STATIC_LOGDIR]}/meminfo"
    cat /proc/cpuinfo > "${cfg[AOD_STATIC_LOGDIR]}/cpuinfo"

    last reboot -5 >> "${cfg[AOD_STATIC_LOGDIR]}/uptime_info"
    cat /proc/uptime >> "${cfg[AOD_STATIC_LOGDIR]}/uptime_info"

    date -u > "${cfg[AOD_STATIC_LOGDIR]}/proc_net_dev"
    cat /proc/net/dev >> "${cfg[AOD_STATIC_LOGDIR]}/proc_net_dev"
    uname -smr > "${cfg[AOD_STATIC_LOGDIR]}/kernel_version"
}

start_tcpdump() {
    if [[ ${cfg[AOD_NW_CAPTURE_ENABLE]} != 0 ]]
    then
        local tcpdump_cmdline="${cfg[AOD_TCPDUMP_BIN]}\
            port ${cfg[AOD_NFS_PORT]}\
            -B ${cfg[AOD_TCPDUMP_BUFSZ_KB]}\
            -w ${cfg[AOD_TCPDUMP_PCAP_PATTERN]}\
            -C ${cfg[AOD_TCPDUMP_PCAPSZ_MB]}\
            -s ${cfg[AOD_TCPDUMP_SNAPLEN_BYTES]}\
            -W ${cfg[AOD_TCPDUMP_MAX_PCAPS]}\
            -n"
        nohup $tcpdump_cmdline > "${cfg[AOD_TCPDUMP_LOGFILE]}" 2>&1 &
        echo $! > "${cfg[AOD_TCPDUMP_PIDFILE]}"
    fi
}

stop_tcpdump() {
    local tcpdump_pid;
    local kill_res;

    if [[ -f "${cfg[AOD_TCPDUMP_PIDFILE]}" ]]
    then
        tcpdump_pid=$(cat "${cfg[AOD_TCPDUMP_PIDFILE]}")

        kill_res=-1
        if (( $(ps --no-headers -p $tcpdump_pid |wc -l) == 0 ))
        then
            # If the tcpdump pidfile is present, but there's no tcpdump process
            # running, we don't want to try and kill the PID.
            kill_res=0
        fi

        while (( kill_res != 0 ))
        do
            kill -9 "$tcpdump_pid";
            kill_res=$?
            sleep 1;
        done

        rm -f "${cfg[AOD_TCPDUMP_PIDFILE]}"
    fi
}

probe_nw_conn() {
    local cur_logdir="$1"
    local srv_ip="$2"
    local srv_port="$3"
    local retval=1
    local logfile="${cur_logdir}/network_probe.out"

    date -u >>$logfile
    nc -z -v -w ${cfg[AOD_NW_PROBE_TIMEOUT_SEC]} $srv_ip $srv_port 2>>$logfile
    retval=$?
    echo >>$logfile

    echo $retval
}

probe_nfs4_mounts() {
    local cur_logdir="$1";
    local storage_acct;
    local nfs4_export;
    local nfs4_mntpt;
    local nfs4_srv_ip;
    local nfs4_srv_port=2049;
    local nfs4_mnt_line;
    local nfs4_mnt_logline;
    local probe_result=1;
    local curdate;
    local probe_err_count=0;
    local logfile="${cur_logdir}/nfs_mounts.out"

    for nfs4_mnt_line in $(mount -t nfs4 |sed -E "${NFS4_MOUNT_LINE_SEDCMD}")
    do
        storage_acct=$(echo $nfs4_mnt_line|cut -d\; -f1);
        nfs4_export=$(echo $nfs4_mnt_line|cut -d\; -f2);
        nfs4_mntpt=$(echo $nfs4_mnt_line|cut -d\; -f3);
        nfs4_srv_ip=$(echo $nfs4_mnt_line|cut -d\; -f4);

        probe_result=$(probe_nw_conn "$cur_logdir" "$nfs4_srv_ip" "$nfs4_srv_port")
        if (( $probe_result != 0 ));
        then
            log_err_to_syslog "network connection probe on $storage_acct -- $nfs4_srv_ip:$nfs4_srv_port failed"
            probe_err_count=$(( probe_err_count + 1 ))
        fi

        nfs4_mnt_logline="storage_acct=$storage_acct export=$nfs4_export mountpt=$nfs4_mntpt srvip=$nfs4_srv_ip"
        echo "[$(date -u)] $nfs4_mnt_logline" >>"$logfile"
    done

    echo $probe_err_count
}

collect_and_tar_logs() {
    local cur_logdir
    local tarball_path
    local nw_probe_err=1

    ##
    # In the staging area, create a directory named with the current log prefix.
    cur_logdir="${cfg[AOD_TMPFS_MNTPT]}/$(mk_logdir_name)"
    mkdir "$cur_logdir"
    if (( $? != 0 ))
    then
        echo "creating logdir ( $cur_logdir ) failed"
        exit 1
    fi

    # dmesg
    journalctl -k -S -"${cfg[AOD_LOG_LOOP_SLEEP_SEC]}sec" > "${cur_logdir}/journalctl.out"

    # network stats; can be used to compute n/w throughput on the interface
    date -u >> "${cur_logdir}/proc_net_dev"
    cat /proc/net/dev >> "${cur_logdir}/proc_net_dev"

    ## nfsstat
    date -u >> "${cur_logdir}/nfsstat.out"
    /usr/sbin/nfsstat -c -l -o nfs 2>&1 |sort -nr -k5 >> "${cur_logdir}/nfsstat.out"

    ## mountstats
    date -u >> "${cur_logdir}/mountstats.out"
    /usr/sbin/mountstats 2>&1 >> "${cur_logdir}/mountstats.out"

    ## mountstats --nfs
    date -u >> "${cur_logdir}/mountstats_nfs.out"
    /usr/sbin/mountstats --nfs 2>&1 >> "${cur_logdir}/mountstats_nfs.out"

    nw_probe_err=$(probe_nfs4_mounts "${cur_logdir}")
    if (( $nw_probe_err != 0 ));
    then
        tarball_path="${cur_logdir}_PROBE_FAILED.tgz"
    else
        tarball_path="${cur_logdir}.tgz"
    fi

    # Bundle up the logdir into a tarball, in the tmpfs partition
    # This works because the tarball_path is prefixed by the tmpfs partition
    # path.
    tar cvzf "$tarball_path" "$cur_logdir" >/dev/null
    if (( $? != 0 ))
    then
        echo "Error creating tarball $tarball_path: $?"
    else
        rm -fR "$cur_logdir"
    fi

    if (( $nw_probe_err != 0 ));
    then
        # continue log collection; but persist the probe failure log.
        log_alert_to_syslog "ALERT -- network probe failed -- ALERT"
        persist_tmpfs_logs
    fi
}

collect_logs_and_prune_tarballs() {
    local tarball_count
    local num_logs_to_delete

    collect_and_tar_logs

    # Preserve only the newest N tarballs
    tarball_count=$(ls -1rt "${cfg[AOD_TMPFS_MNTPT]}"/*aod*.tgz |wc -l)
    if (( tarball_count > ${cfg[AOD_NUM_LOGS_RETAINED]} ));
    then
        num_logs_to_delete=$(( tarball_count - ${cfg[AOD_NUM_LOGS_RETAINED]} ))
        ls -1rt "${cfg[AOD_TMPFS_MNTPT]}"/*aod*.tgz |head -n $num_logs_to_delete |xargs rm -f {}
    fi
}

persist_tmpfs_logs() {
    is_valid_persistent_dir_path
    retval=$?
    if [ "$retval" == 1 ];
    then
        echo "Persisting tmpfs logs"
        find ${cfg[AOD_TMPFS_MNTPT]} -type f -name '*pcap*' -exec cp '{}' ${cfg[AOD_PERSISTENT_LOGDIR]} \;
        find ${cfg[AOD_TMPFS_MNTPT]} -type f -name '*aod*.tgz' -exec cp '{}' ${cfg[AOD_PERSISTENT_LOGDIR]} \;
    fi
}

cleanup() {
    if [[ -f "$AOD_CFG_FILE_STRIPPED" ]];
    then
        rm -f "$AOD_CFG_FILE_STRIPPED"
    fi

    stop_tcpdump
    collect_logs_and_prune_tarballs
    persist_tmpfs_logs
    remove_disk_clean_job
}

log_collection_loop() {
    collect_static_info
    start_tcpdump

    while true;
    do
        if [[ -f "${cfg[AOD_STOP_CAPTURE_MARKER]}" ]];
        then
            cleanup
            rm -f ${cfg[AOD_STOP_CAPTURE_MARKER]}
            break
        fi

        collect_logs_and_prune_tarballs

        sleep ${cfg[AOD_LOG_LOOP_SLEEP_SEC]}
    done
}

# unclean exit, will not be restarted
handle_sigusr1() {
    echo "Received SIGUSR1, stopping"
    date -u > ${cfg[AOD_STOP_CAPTURE_MARKER]}
    exit 1
}

# clean exit, will be restarted
handle_sigterm() {
    echo "Received SIGTERM, stopping"
    date -u > ${cfg[AOD_STOP_CAPTURE_MARKER]}
    cleanup
    exit 0
}

handle_sig0() {
    if (( $(id -u) == 0 ));
    then
        echo "Cleaning up before exiting"
        date -u > "${cfg[AOD_STOP_CAPTURE_MARKER]}";
        cleanup
    fi
}

main() {
    trap handle_sigterm SIGTERM
    trap handle_sigusr1 SIGUSR1
    trap handle_sig0 0

    check_root
    parse_cfg_file
    populate_derived_cfg
    check_persistent_dir_path

    setup_tmp_logdirs
    setup_disk_clean_job

    check_other_running_instances
    log_err_to_syslog "$0: PID is $$"

    log_collection_loop
}

main
